/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is JSIRC Library
 *
 * The Initial Developer of the Original Code is New Dimensions Consulting,
 * Inc. Portions created by New Dimensions Consulting, Inc. are
 * Copyright (C) 1999 New Dimenstions Consulting, Inc. All
 * Rights Reserved.
 *
 * Contributor(s):
 *  Robert Ginda, rginda@ndcico.com, original author
 *  Peter Van der Beken, peter.vanderbeken@pandora.be, necko-only version
 *
 * depends on utils.js, and XPCOM/XPConnect in the JS environment
 *
 */

/* Changes by David Costanzo
   * Make this not depend on utils.js so it can be used independently.
   * Prefix all global symbols with "dict" so that changes to jslib won't
     affect this code (and vice-versa) if it is loaded into the same 
     JavaScript namespace.
*/

var EXPORTED_SYMBOLS = ['DictCBSConnection'];


var DICTJSENV = new Object();
DICTJSENV.HAS_DOCUMENT = (typeof document == "object");
DICTJSENV.HAS_NSPR_EVENTQ = DICTJSENV.HAS_DOCUMENT;
DICTJSENV.HAS_STREAM_PROVIDER = ("nsIStreamProvider" in Components.interfaces);


// HACK:
//   I am changing this to true because the dict protocol handler
//   doesn't work otherwise.  I supsect it should be true, anyway,
//   and that testing if 'document' is defined is not a good way
//   to determine if we have an event queue.
DICTJSENV.HAS_NSPR_EVENTQ = true;

function dictToScriptableInputStream (i)
{
    var si = Components.classes["@mozilla.org/scriptableinputstream;1"];
    
    si = si.createInstance();
    si = si.QueryInterface(Components.interfaces.nsIScriptableInputStream);
    si.init(i);

    return si;
    
}

function DictCBSConnection ()
{
    var sockServiceClass =
        Components.classesByID["{c07e81e0-ef12-11d2-92b6-00105a1b0d64}"];
    
    if (!sockServiceClass)
        throw ("Couldn't get socket service class.");
    
    var sockService = sockServiceClass.getService();
    if (!sockService)
        throw ("Couldn't get socket service.");

    this._sockService = sockService.QueryInterface
        (Components.interfaces.nsISocketTransportService);

    this.wrappedJSObject = this;

}

DictCBSConnection.prototype.connect =
function bc_connect(host, port, bind, tcp_flag, observer)
{
    if (typeof tcp_flag == "undefined")
		tcp_flag = false;
    
    this.host = host.toLowerCase();
    this.port = port;
    this.bind = bind;
    this.tcp_flag = tcp_flag;

    if (DICTJSENV.HAS_STREAM_PROVIDER)
    {
        this._transport = this._sockService.createTransport(host, port, null, 0, 0);

        if (!this._transport)
            throw "Error creating transport.";

        if (DICTJSENV.HAS_NSPR_EVENTQ)
        {
            // we've got an event queue, so start up an async write.
            this._streamProvider = new DictStreamProvider(observer);
            this._write_req = this._transport.asyncWrite(this._streamProvider, this, 0, -1, 0);
        }
        else
        {   /* no nspr event queues in this environment, we can't use async calls,
             * so set up the streams. */
            this._outputStream = this._transport.openOutputStream(0, -1, 0);
            if (!this._outputStream)
                throw "Error getting output stream. (no async, stream provider)";

            this._inputStream = this._transport.openInputStream (0, -1, 0);
            if (!this._inputStream)
                throw "Error getting input stream. (no async, stream provider)";
        }
    }
    else
    {
        // use new necko interfaces (bug 111479)
        this._transport = this._sockService.createTransport(null, 0, host, port, null);
        if (!this._transport)
            throw "Error creating transport.";

        var openFlags;
        if (DICTJSENV.HAS_NSPR_EVENTQ)
            openFlags = 0;
        else
            openFlags = Components.interfaces.nsITransport.OPEN_BLOCKING;

        this._outputStream = this._transport.openOutputStream(openFlags, 4096, 32);
        if (!this._outputStream)
            throw "Error getting output stream. (no stream provider)";

        this._inputStream = this._transport.openInputStream(openFlags, 0, 0);
        if (!this._inputStream)
            throw "Error getting input stream. (no stream provider)";
    }

    this.connectDate = new Date();
    this.isConnected = true;

    return this.isConnected;
  
}

DictCBSConnection.prototype.disconnect =
function bc_disconnect()
{    
    if ("_inputStream" in this && this._inputStream)
        this._inputStream.close();
    if ("_outputStream" in this && this._outputStream)
        this._outputStream.close();

    this.isConnected = false;
}

DictCBSConnection.prototype.sendData =
function bc_senddata(str)
{
    if (!this.isConnected)
        throw "Not Connected.";

    if (DICTJSENV.HAS_NSPR_EVENTQ && DICTJSENV.HAS_STREAM_PROVIDER)
        this.asyncWrite(str);
    else
        this.sendDataNow(str);
}
    
DictCBSConnection.prototype.readData =
function bc_readdata(timeout, count)
{
    if (!this.isConnected)
        throw "Not Connected.";

    var rv;

    try
    {
        rv = this._scriptableInputStream.read(count);
    }
    catch (ex)
    {
        //dd("*** Caught " + ex + " while reading.")
        this.isConnected = false;
        throw ex;
    }
    
    return rv;
}

DictCBSConnection.prototype.startAsyncRead =
function bc_saread (observer)
{
    if (DICTJSENV.HAS_STREAM_PROVIDER)
    {
        this._transport.asyncRead(new DictStreamListener(observer), this, 0, -1, 0);
    }
    else
    {
        var cls = Components.classes["@mozilla.org/network/input-stream-pump;1"];
        var pump = cls.createInstance(Components.interfaces.nsIInputStreamPump);
        pump.init(this._inputStream, -1, -1, 0, 0, false);
        pump.asyncRead(new DictStreamListener(observer), this);
    }

}

DictCBSConnection.prototype.asyncWrite =
function bc_awrite (str)
{
    this._streamProvider.pendingData += str;
    if (this._streamProvider.isBlocked)
    {
        this._write_req.resume();
        this._streamProvider.isBlocked = false;
    }
}

DictCBSConnection.prototype.hasPendingWrite =
function bc_haspwrite ()
{
    if (DICTJSENV.HAS_STREAM_PROVIDER)
        return (this._streamProvider.pendingData != "");
    else
        return false; // data already pushed to necko
}

DictCBSConnection.prototype.sendDataNow =
function bc_senddatanow(str)
{
    var rv = false;
    
    try
    {
        this._outputStream.write(str, str.length);
        rv = true;
    }
    catch (ex)
    {
        //dd ("*** Caught " + ex + " while sending.")
        this.isConnected = false;
        throw (ex);
    }
    
    return rv;
}

function _notimpl ()
{
    throw "Not Implemented.";
}

if (!DICTJSENV.HAS_NSPR_EVENTQ)
{
    DictCBSConnection.prototype.startAsyncRead = _notimpl;
    DictCBSConnection.prototype.asyncWrite = _notimpl;
}
else if (DICTJSENV.HAS_STREAM_PROVIDER)
{
    DictCBSConnection.prototype.sendDataNow = _notimpl;
}
else
{
    DictCBSConnection.prototype.asyncWrite = _notimpl;
}

delete _notimpl;


    
function DictStreamProvider(observer)
{
    this._observer = observer;
}

DictStreamProvider.prototype.pendingData = "";
DictStreamProvider.prototype.isBlocked = true;

DictStreamProvider.prototype.onDataWritable =
function sp_datawrite (request, ctxt, ostream, offset, count)
{
    if ("isClosed" in this && this.isClosed)
        throw Components.results.NS_BASE_STREAM_CLOSED;

    if (!this.pendingData)
    {
        this.isBlocked = true;

        /* this is here to support pre-XPCDOM builds (0.9.0 era), which
         * don't have this result code mapped. */
        if (!Components.results.NS_BASE_STREAM_WOULD_BLOCK)
            throw 2152136711;
        
        throw Components.results.NS_BASE_STREAM_WOULD_BLOCK;
    }
    
    var len = ostream.write (this.pendingData, this.pendingData.length);
    this.pendingData = this.pendingData.substr (len);
}

DictStreamProvider.prototype.onStartRequest =
function sp_startreq (request, ctxt)
{
}


DictStreamProvider.prototype.onStopRequest =
function sp_stopreq (request, ctxt, status)
{
    if (this._observer)
        this._observer.onStreamClose(status);
}

function DictStreamListener(observer)
{
    this._observer = observer;
}

DictStreamListener.prototype.onStartRequest =
function sl_startreq (request, ctxt)
{
}

DictStreamListener.prototype.onStopRequest =
function sl_stopreq (request, ctxt, status)
{
    if (this._observer)
        this._observer.onStreamClose(status);
}

DictStreamListener.prototype.onDataAvailable =
function sl_dataavail (request, ctxt, inStr, sourceOffset, count)
{
    ctxt = ctxt.wrappedJSObject;
    if (!ctxt)
    {
        //dd ("*** Can't get wrappedJSObject from ctxt in " +
        //    "DictStreamListener.onDataAvailable ***");
        return;
    }

    if (!("_scriptableInputStream" in ctxt))
        ctxt._scriptableInputStream = dictToScriptableInputStream(inStr);

    if (this._observer)
        this._observer.onStreamDataAvailable(request, inStr, sourceOffset, count);
}

